@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: mouse.c,v 1.117.2.4 2010/08/06 01:23:06 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: mouse.c,v 1.133.2.2 2012/09/26 23:05:56 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - mouse.c */
@@ -71,6 +71,7 @@ static char *RCSid() { return RCSid("$Id: mouse.c,v 1.117.2.4 2010/08/06 01:23:0
 #endif
 
 #ifdef OS2
+#include <os2.h>
 #include "os2/pm_msgs.h"
 #endif
 
@@ -257,7 +258,7 @@ static void put_label __PROTO((char *label, double x, double y));
 static void
 alert()
 {
-# ifdef GNUPMDRV
+# ifdef OS2
     DosBeep(444, 111);
 # else
 #  ifdef HAVE_LIBREADLINE
@@ -371,22 +372,22 @@ MousePosToGraphPosReal(int xx, int yy, double *x, double *y, double *x2, double
 static char *
 xy_format()
 {
-    static char format[0xff];
+    static char format[64];
     format[0] = NUL;
-    strcat(format, mouse_setting.fmt);
-    strcat(format, ", ");
-    strcat(format, mouse_setting.fmt);
+    strncat(format, mouse_setting.fmt, 30);
+    strncat(format, ", ", 2);
+    strncat(format, mouse_setting.fmt, 30);
     return format;
 }
 
 static char *
 zoombox_format()
 {
-    static char format[0xff];
+    static char format[64];
     format[0] = NUL;
-    strcat(format, mouse_setting.fmt);
-    strcat(format, "\r");
-    strcat(format, mouse_setting.fmt);
+    strncat(format, mouse_setting.fmt, 30);
+    strncat(format, "\r", 2);
+    strncat(format, mouse_setting.fmt, 30);
     return format;
 }
 
@@ -424,8 +425,25 @@ GetAnnotateString(char *s, double x, double y, int mode, char *fmt)
 	}
     } else if (mode == MOUSE_COORDINATES_REAL1) {
 	sprintf(s, xy_format(), x, y);	/* w/o brackets */
-    } else if (mode == MOUSE_COORDINATES_ALT && fmt) {
-	sprintf(s, fmt, x, y);	/* user defined format */
+    } else if (mode == MOUSE_COORDINATES_ALT && (fmt || polar)) {
+	if (polar) {
+	    double phi, r;
+	    double rmin = (R_AXIS.autoscale & AUTOSCALE_MIN) ? 0.0 : R_AXIS.set_min;
+	    phi = atan2(y,x);
+	    if (R_AXIS.log)
+		r = AXIS_UNDO_LOG(POLAR_AXIS, x/cos(phi) + AXIS_DO_LOG(POLAR_AXIS, rmin));
+	    else
+		r = x/cos(phi) + rmin;
+	    if (fmt)
+		sprintf(s, fmt, phi/ang2rad, r);
+	    else {
+		sprintf(s, "polar: ");
+		s += strlen(s);
+		sprintf(s, xy_format(), phi/ang2rad, r);
+	    }
+	} else {
+	    sprintf(s, fmt, x, y);	/* user defined format */
+	}
     } else {
 	sprintf(s, xy_format(), x, y);	/* usual x,y values */
     }
@@ -487,7 +505,7 @@ xDateTimeFormat(double x, char *b, int mode)
 #define MKSTR(sp,x,axis)					\
 do {								\
     if (x >= VERYLARGE)	break;					\
-    if (axis_array[axis].is_timedata) { 			\
+    if (axis_array[axis].datatype == DT_TIMEDATE) {		\
 	char *format = copy_or_invent_formatstring(axis);	\
 	while (strchr(format,'\n'))				\
 	     *(strchr(format,'\n')) = ' ';			\
@@ -622,6 +640,9 @@ apply_zoom(struct t_zoom *z)
 	return;
     }
 
+    /* Now we're committed. Notify the terminal the the next replot is a zoom */
+    (*term->layer)(TERM_LAYER_BEFORE_ZOOM);
+
     flip = (is_splot_map && zoom_now->was_splot_map);
     sprintf(s, "set xr[%.12g:%.12g]; set yr[%.12g:%.12g]",
 	       zoom_now->xmin, zoom_now->xmax, 
@@ -643,6 +664,7 @@ apply_zoom(struct t_zoom *z)
 	for (i=0; i<AXIS_ARRAY_SIZE; i++) {
 	    axis_array_copy[i].label = axis_array[i].label;
 	    axis_array_copy[i].ticdef.def.user = axis_array[i].ticdef.def.user;
+	    axis_array_copy[i].ticdef.font = axis_array[i].ticdef.font;
 	}
 	memcpy(axis_array, axis_array_copy, sizeof(axis_array));
 	s[0] = '\0';	/* FIXME:  Is this better than calling replotrequest()? */
@@ -661,9 +683,12 @@ apply_zoom(struct t_zoom *z)
 	}
 #endif
 
+    } else {
+	inside_zoom = TRUE;
     }
 
     do_string_replot(s);
+    inside_zoom = FALSE;
 }
 
 
@@ -743,21 +768,18 @@ incr_mousemode(const int amount)
 {
     long int old = mouse_mode;
     mouse_mode += amount;
-    if (MOUSE_COORDINATES_ALT == mouse_mode && !mouse_alt_string) {
+    if (MOUSE_COORDINATES_ALT == mouse_mode && !(mouse_alt_string || polar))
 	mouse_mode += amount;	/* stepping over */
-    }
     if (mouse_mode > MOUSE_COORDINATES_ALT) {
-	mouse_mode = MOUSE_COORDINATES_REAL;
-    } else if (mouse_mode < MOUSE_COORDINATES_REAL) {
+	mouse_mode = MOUSE_COORDINATES_REAL1;
+    } else if (mouse_mode <= MOUSE_COORDINATES_REAL) {
 	mouse_mode = MOUSE_COORDINATES_ALT;
-	if (MOUSE_COORDINATES_ALT == mouse_mode && !mouse_alt_string) {
+	if (!(mouse_alt_string || polar))
 	    mouse_mode--;	/* stepping over */
-	}
     }
     UpdateStatusline();
-    if (display_ipc_commands()) {
+    if (display_ipc_commands())
 	fprintf(stderr, "switched mouse format from %ld to %ld\n", old, mouse_mode);
-    }
 }
 
 static void
@@ -1003,13 +1025,13 @@ builtin_nearest_log(struct gp_event_t *ge)
 	    change_y2 = TRUE;
 	
 	if (change_x1)
-	    do_string(axis_array[FIRST_X_AXIS].log ? "unset log x" : "set log x", FALSE);
+	    do_string(axis_array[FIRST_X_AXIS].log ? "unset log x" : "set log x");
 	if (change_y1)
-	    do_string(axis_array[FIRST_Y_AXIS].log ? "unset log y" : "set log y", FALSE);
+	    do_string(axis_array[FIRST_Y_AXIS].log ? "unset log y" : "set log y");
 	if (change_x2 && !splot_map)
-	    do_string(axis_array[SECOND_X_AXIS].log ? "unset log x2" : "set log x2", FALSE);
+	    do_string(axis_array[SECOND_X_AXIS].log ? "unset log x2" : "set log x2");
 	if (change_y2 && !splot_map)
-	    do_string(axis_array[SECOND_Y_AXIS].log ? "unset log y2" : "set log y2", FALSE);
+	    do_string(axis_array[SECOND_Y_AXIS].log ? "unset log y2" : "set log y2");
 	if (!change_x1 && !change_y1 && splot_map)
 	    do_string_replot( Z_AXIS.log ? "unset log z" : "set log z");
 	
@@ -1091,7 +1113,7 @@ static char *
 builtin_decrement_mousemode(struct gp_event_t *ge)
 {
     if (!ge) {
-	return "`builtin-decrement-mousemode`";
+	return "`builtin-previous-mouse-format`";
     }
     incr_mousemode(-1);
     return (char *) 0;
@@ -1101,7 +1123,7 @@ static char *
 builtin_increment_mousemode(struct gp_event_t *ge)
 {
     if (!ge) {
-	return "`builtin-increment-mousemode`";
+	return "`builtin-next-mouse-format`";
     }
     incr_mousemode(1);
     return (char *) 0;
@@ -1287,7 +1309,7 @@ event_keypress(struct gp_event_t *ge, TBOOLEAN current)
 	bind_install_default_bindings();
     }
 
-    if (modifier_mask & Mod_Shift) {
+    if ((modifier_mask & Mod_Shift) && ((c & 0xff) == 0)) {
 	c = toupper(c);
     }
 
@@ -1331,7 +1353,7 @@ event_keypress(struct gp_event_t *ge, TBOOLEAN current)
 		    /* FIXME - Better to clear MOUSE_[XY] than to set it wrongly. */
 		    /*         This may be worth a separate subroutine.           */
 		    load_mouse_variables(0, 0, FALSE, c);
-		do_string(ptr->command, FALSE);
+		do_string(ptr->command);
 		/* Treat as a current event after we return to x11.trm */
 		ge->type = GE_keypress;
 		break;
@@ -1340,7 +1362,7 @@ event_keypress(struct gp_event_t *ge, TBOOLEAN current)
 		break;
 	    /* Let user defined bindings overwrite the builtin bindings */
 	    } else if ((par2 & 1) == 0 && ptr->command) {
-		do_string(ptr->command, FALSE);
+		do_string(ptr->command);
 		break;
 	    } else if (ptr->builtin) {
 		ptr->builtin(ge);
@@ -1395,6 +1417,16 @@ ChangeView(int x, int z)
     }
 }
 
+/* Return a new (upper or lower) axis limit that is a linear
+   combination of the current limits.  */
+static double
+rescale(int AXIS, double w1, double w2) 
+{
+  double logval, val;
+  logval = w1*axis_array[AXIS].min+w2*axis_array[AXIS].max;
+  val = AXIS_DE_LOG_VALUE(AXIS,logval);
+  return val;
+}
 
 static void
 event_buttonpress(struct gp_event_t *ge)
@@ -1413,10 +1445,183 @@ event_buttonpress(struct gp_event_t *ge)
 
     MousePosToGraphPosReal(mouse_x, mouse_y, &real_x, &real_y, &real_x2, &real_y2);
 
-    if (ALMOST2D) {
-	if (!setting_zoom_region) {
+
+    if ((b == 4 || b == 6) && /* 4 - wheel up, 6 - wheel left */
+	(!replot_disabled || refresh_ok)	/* Use refresh if available */
+	&& !(paused_for_mouse & PAUSE_BUTTON3)) {
+      double xmin, ymin, x2min, y2min;
+      double xmax, ymax, x2max, y2max;
+      if ((b == 4) && (modifier_mask & Mod_Ctrl)) {
+	if (modifier_mask & Mod_Shift) {
+	  /* zoom in (X axis only) */
+	  double w1=23./25.;
+	  double w2= 2./25.;
+	  xmin = rescale(FIRST_X_AXIS, w1, w2);
+	  ymin = rescale(FIRST_Y_AXIS, 1., 0.);
+	  x2min = rescale(SECOND_X_AXIS, w1, w2);
+	  y2min = rescale(SECOND_Y_AXIS, 1., 0.);
+
+	  xmax = rescale(FIRST_X_AXIS, w2,w1);
+	  ymax = rescale(FIRST_Y_AXIS, 0., 1.);
+	  x2max = rescale(SECOND_X_AXIS, w2, w1);
+	  y2max = rescale(SECOND_Y_AXIS, 0., 1.);
+
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "zoom in X.\n");
+	  }
+	} else {
+	  /* zoom in (factor of approximatly 2^(.25), so four steps gives 2x larger) */
+	  double w1=23./25.;
+	  double w2= 2./25.;
+	  xmin = rescale(FIRST_X_AXIS,  w1, w2);
+	  ymin = rescale(FIRST_Y_AXIS,  w1, w2);
+	  x2min = rescale(SECOND_X_AXIS, w1, w2);
+	  y2min = rescale(SECOND_Y_AXIS, w1, w2);
+
+	  xmax = rescale(FIRST_X_AXIS,  w2, w1);
+	  ymax = rescale(FIRST_Y_AXIS,  w2, w1);
+	  x2max = rescale(SECOND_X_AXIS, w2, w1);
+	  y2max = rescale(SECOND_Y_AXIS, w2, w1);
+
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "zoom in.\n");
+	  }
+	}
+      } else {
+#if DO_ZOOM_AND_APPLY_ZOOM_FIXED_TO_HANDLE_Z_AXIS
+	if ((b == 4) && (modifier_mask & Mod_Alt)) {
+	  /* scroll in +Z direction */
+	  xmin = rescale(FIRST_X_AXIS, 1., 0.);
+	  ymin = rescale(FIRST_Y_AXIS, 1., 0.);
+	  x2min = rescale(SECOND_X_AXIS, 1., 0.);
+	  y2min = rescale(SECOND_Y_AXIS, 1., 0.);
+	  zmin = rescale(FIRST_Z_AXIS, .9, .1);
+	  
+	  xmax = rescale(FIRST_X_AXIS,  0., 1.);
+	  ymax = rescale(FIRST_Y_AXIS,  0., 1.);
+	  x2max = rescale(SECOND_X_AXIS, 0., 1.);
+	  y2max = rescale(SECOND_Y_AXIS, 0., 1.);
+	  zmax = rescale(FIRST_Z_AXIS, 1.1, -.1);
+	  do_zoom(xmin, ymin, x2min, y2min, zmin, xmax, ymax, x2max, y2max, zmax);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "scroll left.\n");
+	  }
+	} else 
+#endif
+	  if (((b == 4) && (modifier_mask & Mod_Shift)) || (b == 6)) {
+	  /* scroll left */
+          xmin = rescale(FIRST_X_AXIS, 1.1, -.1);
+	  ymin = rescale(FIRST_Y_AXIS, 1., 0.);
+	  x2min = rescale(SECOND_X_AXIS, 1.1, -.1);
+	  y2min = rescale(SECOND_Y_AXIS, 1., 0.);
+	  
+	  xmax = rescale(FIRST_X_AXIS, .1, .9);
+	  ymax = rescale(FIRST_Y_AXIS, 0., 1.);
+	  x2max = rescale(SECOND_X_AXIS, .1, .9);
+	  y2max = rescale(SECOND_Y_AXIS, 0., 1.);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "scroll left.\n");
+	  }
+	} else {
+	  /* scroll up */
+	  xmin = rescale(FIRST_X_AXIS, 1., 0.);
+	  ymin = rescale(FIRST_Y_AXIS, .9,.1);
+	  x2min = rescale(SECOND_X_AXIS, 1.,0.);
+	  y2min = rescale(SECOND_Y_AXIS, .9, .1);
+
+	  xmax = rescale(FIRST_X_AXIS, 0., 1.);
+	  ymax = rescale(FIRST_Y_AXIS, -.1,1.1);
+	  x2max = rescale(SECOND_X_AXIS, 0., 1.);
+	  y2max = rescale(SECOND_Y_AXIS, -.1, 1.1);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "scroll up.\n");
+	  }
+	}
+      }
+    } else if (((b == 5) || (b == 7)) && /* 5 - wheel down, 7 - wheel right */
+	       (!replot_disabled || refresh_ok)	/* Use refresh if available */
+	       && !(paused_for_mouse & PAUSE_BUTTON3)) {
+      double xmin, ymin, x2min, y2min;
+      double xmax, ymax, x2max, y2max;
+      if ((b == 5) && (modifier_mask & Mod_Ctrl)) {
+	if (modifier_mask & Mod_Shift) {
+	  /* zoom out (X axis only) */
+	  double w1= 23./21.;
+	  double w2=-2./21.;
+	  xmin = rescale(FIRST_X_AXIS, w1, w2);
+	  ymin = rescale(FIRST_Y_AXIS, 1., 0.);
+	  x2min = rescale(SECOND_X_AXIS, w1, w2);
+	  y2min = rescale(SECOND_Y_AXIS, 1., 0.);
+	  
+	  xmax = rescale(FIRST_X_AXIS, w2, w1);
+	  ymax = rescale(FIRST_Y_AXIS, 0., 1.);
+	  x2max = rescale(SECOND_X_AXIS, w2, w1);
+	  y2max = rescale(SECOND_Y_AXIS, 0., 1.);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "zoom out X.\n");
+	  }
+	} else {
+	  /* zoom out (exactly undo a zoom in step) */
+	  double w1= 23./21.;
+	  double w2=-2./21.;
+	  xmin = rescale(FIRST_X_AXIS,  w1, w2);
+	  ymin = rescale(FIRST_Y_AXIS,  w1, w2);
+	  x2min = rescale(SECOND_X_AXIS, w1, w2);
+	  y2min = rescale(SECOND_Y_AXIS, w1, w2);
+
+	  xmax = rescale(FIRST_X_AXIS,  w2, w1);
+	  ymax = rescale(FIRST_Y_AXIS,  w2, w1);
+	  x2max = rescale(SECOND_X_AXIS, w2, w1);
+	  y2max = rescale(SECOND_Y_AXIS, w2, w1);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "zoom out.\n");
+	  }
+	}
+      } else {
+	if (((b == 5) && (modifier_mask & Mod_Shift)) || (b == 7)) {
+	  /* scroll right */
+	  xmin = rescale(FIRST_X_AXIS, .9, .1);
+	  ymin = rescale(FIRST_Y_AXIS, 1., 0.);
+	  x2min = rescale(SECOND_X_AXIS, .9, .1);
+	  y2min = rescale(SECOND_Y_AXIS, 1., 0.);
+
+	  xmax = rescale(FIRST_X_AXIS, -.1, 1.1);
+	  ymax = rescale(FIRST_Y_AXIS, 0., 1.);
+	  x2max = rescale(SECOND_X_AXIS, -.1, 1.1);
+	  y2max = rescale(SECOND_Y_AXIS, 0., 1.);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "scroll right.\n");
+	  }
+	} else {
+	  /* scroll down */
+	  xmin = rescale(FIRST_X_AXIS, 1., 0.);
+	  ymin = rescale(FIRST_Y_AXIS, 1.1, -.1);
+	  x2min = rescale(SECOND_X_AXIS, 1., 0.);
+	  y2min = rescale(SECOND_Y_AXIS, 1.1, -.1);
+
+	  xmax = rescale(FIRST_X_AXIS, 0., 1.);
+	  ymax = rescale(FIRST_Y_AXIS, .1, .9);
+	  x2max = rescale(SECOND_X_AXIS, 0., 1.);
+	  y2max = rescale(SECOND_Y_AXIS, .1, .9);
+	  rescale(FIRST_Y_AXIS, 1.1,-.1);
+	  rescale(FIRST_Y_AXIS, .1,.9);
+	  do_zoom(xmin, ymin, x2min, y2min, xmax, ymax, x2max, y2max);
+	  if (display_ipc_commands()) {
+	    fprintf(stderr, "scroll down.\n");
+	  }
+	}
+      }
+    } else if (ALMOST2D) {
+        if (!setting_zoom_region) {
 	    if (1 == b) {
-		/* not bound in 2d graphs */
+	      /* not bound in 2d graphs */
 	    } else if (2 == b) {
 		/* not bound in 2d graphs */
 	    } else if (3 == b && 
@@ -1681,7 +1886,8 @@ event_motion(struct gp_event_t *ge)
 	    } else {
 
 		if (relx > rely) {
-		    surface_scale += (mouse_x - start_x) * 2.0 / term->xmax;
+		    surface_lscale += (mouse_x - start_x) * 2.0 / term->xmax;
+		    surface_scale = exp(surface_lscale);
 		    if (surface_scale < 0)
 			surface_scale = 0;
 		} else {
@@ -1856,7 +2062,7 @@ do_event(struct gp_event_t *ge)
 	break;
     case GE_replot:
 	/* used only by ggi.trm */
-	do_string("replot", FALSE);
+	do_string("replot");
 	break;
     case GE_reset:
 	event_reset(ge);
@@ -2090,6 +2296,7 @@ bind_display(char *lhs)
 	/* display all bindings */
 	char fmt[] = " %-17s  %s\n";
 	fprintf(stderr, "\n");
+	/* mouse buttons */
 	fprintf(stderr, fmt, "2x<B1>",
 		"print coordinates to clipboard using `clipboardformat`\n                    (see keys '3', '4')");
 	fprintf(stderr, fmt, "<B2>", "annotate the graph using `mouseformat` (see keys '1', '2')");
@@ -2099,7 +2306,18 @@ bind_display(char *lhs)
 	fprintf(stderr, fmt, "<B1-Motion>", "change view (rotation). Use <ctrl> to rotate the axes only.");
 	fprintf(stderr, fmt, "<B2-Motion>", "change view (scaling). Use <ctrl> to scale the axes only.");
 	fprintf(stderr, fmt, "<Shift-B2-Motion>", "vertical motion -- change xyplane");
+	/* mouse wheel */
+	fprintf(stderr, fmt, "<wheel-up>", "scroll up (in +Y direction).");
+	fprintf(stderr, fmt, "<wheel-down>", "scroll down.");
+	fprintf(stderr, fmt, "<shift-wheel-up>", "scroll left (in -X direction).");
+	fprintf(stderr, fmt, "<shift-wheel-down>", "scroll right.");
+	fprintf(stderr, fmt, "<control-wheel-up>", "zoom in toward the center of the plot.");
+	fprintf(stderr, fmt, "<control-wheel-down>", " zoom out.");
+	fprintf(stderr, fmt, "<shift-control-wheel-up>", "zoom in only the X axis.");
+	fprintf(stderr, fmt, "<shift-control-wheel-down>", "zoom out only the X axis.");
+
 	fprintf(stderr, "\n");
+	/* keystrokes */
 #ifndef DISABLE_SPACE_RAISES_CONSOLE
 	fprintf(stderr, " %-12s   %s\n", "Space", "raise gnuplot console window");
 #endif
