@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: tabulate.c,v 1.7.2.2 2009/12/25 00:23:24 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: tabulate.c,v 1.13.2.2 2012/09/05 23:35:46 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - tabulate.c */
@@ -45,6 +45,7 @@ static char *RCSid() { return RCSid("$Id: tabulate.c,v 1.7.2.2 2009/12/25 00:23:
 
 #include "alloc.h"
 #include "axis.h"
+#include "datafile.h"
 #include "gp_time.h"
 #include "graphics.h"
 #include "graph3d.h"
@@ -62,7 +63,16 @@ static FILE *outfile;
 
 static void
 output_number(double coord, int axis, char *buffer) {
-    if (axis_array[axis].is_timedata) {
+    /* treat timedata and "%s" output format as a special case:
+     * return a number.
+     * "%s" in combination with any other character is treated
+     * like a normal time format specifier and handled in time.c
+     */
+    if (axis_array[axis].datatype == DT_TIMEDATE &&
+	strcmp(axis_array[axis].formatstring, "%s") == 0) {
+	gprintf(buffer, BUFFERSIZE, "%.0f", 1.0, coord);
+    } else
+    if (axis_array[axis].datatype == DT_TIMEDATE) {
 	buffer[0] = '"';
 	gstrftime(buffer+1, BUFFERSIZE-1, axis_array[axis].formatstring, coord);
 	while (strchr(buffer,'\n')) {*(strchr(buffer,'\n')) = ' ';}
@@ -116,9 +126,11 @@ print_table(struct curve_points *current_plot, int plot_num)
 	    fputs("1 y2", outfile);
 	    break;
 	case FINANCEBARS:
-	case CANDLESTICKS:
 	    fputs(" open ylow yhigh yclose", outfile);
 	    break;
+	case CANDLESTICKS:
+	    fputs(" open ylow yhigh yclose width", outfile);
+	    break;
 	case LABELPOINTS:
 	    fputs(" label",outfile);
 	    break;
@@ -134,16 +146,25 @@ print_table(struct curve_points *current_plot, int plot_num)
 	case FSTEPS:
 	case HISTEPS:
 	    break;
+	case IMAGE:
+	    fputs("  pixel", outfile);
+	    break;
+	case RGBIMAGE:
+	case RGBA_IMAGE:
+	    fputs("  red green blue alpha", outfile);
+	    break;
+
 	default:
 	    if (interactive)
-	        fprintf(stderr, "Tabular output of %s plot style not fully implemented\n",
+		fprintf(stderr, "Tabular output of %s plot style not fully implemented\n",
 		    current_plot->plot_style == HISTOGRAMS ? "histograms" :
-		    current_plot->plot_style == IMAGE ? "image" :
-		    current_plot->plot_style == RGBIMAGE ? "rgbimage" :
 		    "this");
 	    break;
 	}
 
+	if (current_plot->varcolor)
+	    fputs("  color", outfile);
+
 	fputs(" type\n", outfile);
 
 	if (current_plot->plot_style == LABELPOINTS) {
@@ -165,6 +186,12 @@ print_table(struct curve_points *current_plot, int plot_num)
 	    for (i = 0, point = current_plot->points; i < current_plot->p_count;
 		i++, point++) {
 
+		/* Reproduce blank lines read from original input file, if any */
+		if (!memcmp(point, &blank_data_line, sizeof(struct coordinate))) {
+		    fprintf(outfile, "\n");
+		    continue;
+		}
+
 		/* FIXME HBB 20020405: had better use the real x/x2 axes of this plot */
 		OUTPUT_NUMBER(point->x, current_plot->x_axis);
 		OUTPUT_NUMBER(point->y, current_plot->y_axis);
@@ -187,14 +214,29 @@ print_table(struct curve_points *current_plot, int plot_num)
 			OUTPUT_NUMBER(point->ylow, current_plot->y_axis);
 			OUTPUT_NUMBER(point->yhigh, current_plot->y_axis);
 			break;
+		    case IMAGE:
+			fprintf(outfile,"%g ",point->z);
+			break;
+		    case RGBIMAGE:
+		    case RGBA_IMAGE:
+			fprintf(outfile,"%4d ",(int)point->CRD_R);
+			fprintf(outfile,"%4d ",(int)point->CRD_G);
+			fprintf(outfile,"%4d ",(int)point->CRD_B);
+			fprintf(outfile,"%4d ",(int)point->CRD_A);
+			break;
 		    case FILLEDCURVES:
 			OUTPUT_NUMBER(point->yhigh, current_plot->y_axis);
 			break;
 		    case FINANCEBARS:
+			OUTPUT_NUMBER(point->ylow, current_plot->y_axis);
+			OUTPUT_NUMBER(point->yhigh, current_plot->y_axis);
+			OUTPUT_NUMBER(point->z, current_plot->y_axis);
+			break;
 		    case CANDLESTICKS:
 			OUTPUT_NUMBER(point->ylow, current_plot->y_axis);
 			OUTPUT_NUMBER(point->yhigh, current_plot->y_axis);
 			OUTPUT_NUMBER(point->z, current_plot->y_axis);
+			OUTPUT_NUMBER(2. * (point->x - point->xlow), current_plot->x_axis);
 			break;
 		    case VECTOR:
 			OUTPUT_NUMBER((point->xhigh - point->x), current_plot->x_axis);
@@ -213,6 +255,19 @@ print_table(struct curve_points *current_plot, int plot_num)
 			/* ? */
 			break;
 		} /* switch(plot type) */
+
+		if (current_plot->varcolor) {
+		    double colorval = current_plot->varcolor[i];
+		    if ((current_plot->lp_properties.pm3d_color.value < 0.0)
+		    &&  (current_plot->lp_properties.pm3d_color.type == TC_RGB)) {
+			fprintf(outfile, "0x%6x", (unsigned int)(colorval));
+		    } else if (current_plot->lp_properties.pm3d_color.type == TC_Z) {
+			OUTPUT_NUMBER(colorval, COLOR_AXIS);
+		    } else if (current_plot->lp_properties.l_type == LT_COLORFROMCOLUMN) {
+			OUTPUT_NUMBER(colorval, COLOR_AXIS);
+		    }
+		}
+
 		fprintf(outfile, " %c\n",
 		    current_plot->points[i].type == INRANGE
 		    ? 'i' : current_plot->points[i].type == OUTRANGE
@@ -249,7 +304,9 @@ print_3dtable(int pcount)
 	    free(title);
 	}
 
-	if (this_plot->plot_style == LABELPOINTS) {
+	switch (this_plot->plot_style) {
+	case LABELPOINTS:
+	    {
 	    struct text_label *this_label;
 	    for (this_label = this_plot->labels->next; this_label != NULL;
 		 this_label = this_label->next) {
@@ -260,6 +317,17 @@ print_3dtable(int pcount)
 		fprintf(outfile, " \"%s\"\n", label);
 		free(label);
 	    }
+	    }
+	    continue;
+	case LINES:
+	case POINTSTYLE:
+	case IMPULSES:
+	case DOTS:
+	case VECTOR:
+	case IMAGE:
+	    break;
+	default:
+	    fprintf(stderr, "Tabular output of this 3D plot style not implemented\n");
 	    continue;
 	}
 
@@ -292,6 +360,8 @@ print_3dtable(int pcount)
 			OUTPUT_NUMBER((tail->y - point->y), FIRST_Y_AXIS);
 			OUTPUT_NUMBER((tail->z - point->z), FIRST_Z_AXIS);
 			tail++;
+		    } else if (this_plot->plot_style == IMAGE) {
+			fprintf(outfile,"%g ",point->CRD_COLOR);
 		    }
 		    fprintf(outfile, "%c\n",
 			    point->type == INRANGE
